from cython.parallel cimport prange, parallel
from libc.stdlib cimport malloc, free
cimport openmp
import cython
import numpy as np

cpdef int lamda_sign(unsigned int lambda_, unsigned int dim) nogil:
    # Returns 1 if lambda_ has an even number of zeros in its binary expansion and -1 otherwise
    # dim is the length of the binary representation of lambda_
    # This sign is necessary for computation of the central difference
    cdef int ones_count = 0
    while lambda_:
        ones_count += lambda_ & 1
        lambda_ >>= 1
    return 1 if (dim - ones_count) % 2 == 0 else -1

cpdef double[:] get_alpha(double[:] beta, unsigned int scaling_depth, unsigned int dim):
    # Returns Vandermonde coefficients alpha given beta
    cdef double[:] alpha_array = np.empty(scaling_depth + 1, dtype = np.double)
    cdef int sign = -1 if scaling_depth & 1 else 1 #(-1) ** scaling_depth
    cdef double alpha
    cdef int i

    for i in range(scaling_depth + 1):
        alpha = sign
        alpha /= beta[i] ** dim
        for j in range(scaling_depth + 1):
            if j == i:
                continue
            alpha *= beta[j] / (beta[i] - beta[j])
        alpha_array[i] = alpha
    return alpha_array

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef double get_coeff_cython(double[:,:] path, unsigned int scaling_depth, unsigned int dyadic_order_1, unsigned int dyadic_order_2, bint use_parallel):

    cdef unsigned int dim = path.shape[1]

    # In the method, lambda represents a component-wise scaling. In the case of a central difference,
    # lambda is a vector in {-1,1}^dim. We choose to represent this vector as an integer, whose binary
    # representation denotes the entries of the vector. For example, 5 = (1,0,1)_2 corresponds to the
    # vector (1, -1, 1). max_lambda here denotes the upper bound for these integers.
    cdef int max_lambda = 1 << dim

    cdef double result = 0
    cdef int lambda_, sgn, i, max_threads, chunk

    cdef double[:] beta = np.linspace(0.1, 1, scaling_depth + 1, dtype = np.double) if scaling_depth > 0 else np.array([1.], dtype = np.double)
    cdef double[:] alpha = get_alpha(beta, scaling_depth, dim)

    if use_parallel:
        max_threads = openmp.omp_get_max_threads()
        openmp.omp_set_num_threads(max_threads)
        chunk = <int>(max_lambda / max_threads)

        for lambda_ in prange(max_lambda, num_threads=max_threads, chunksize = chunk, schedule ="static", nogil=True):
            sgn = lamda_sign(lambda_, dim)
            for i in range(scaling_depth + 1):
                result += sgn * alpha[i] * (get_kernel_cython(path, lambda_, beta[i], dyadic_order_1, dyadic_order_2) - 1)

    else:
        for lambda_ in range(max_lambda):
            sgn = lamda_sign(lambda_, dim)
            for i in range(scaling_depth+1):
                result += sgn * alpha[i] * (get_kernel_cython(path, lambda_, beta[i], dyadic_order_1, dyadic_order_2) - 1)

    result *= 0.5 ** dim
    return result

@cython.boundscheck(False)
@cython.wraparound(False)
cdef double get_kernel_cython(double[:,:] path, int lambda_, double beta, unsigned int dyadic_order_1, unsigned int dyadic_order_2) nogil:
    cdef int path_length = path.shape[0]
    cdef int dim = path.shape[1]
    cdef int i, j, sgn, d, ii
    cdef double beta_frac = beta / (1 << (dyadic_order_1 + dyadic_order_2))
    cdef double twelth = 1. / 12
    cdef double result

    # Dyadically refined grid dimensions
    cdef int dyadic_len = ((path_length - 1) << dyadic_order_1) + 1
    cdef int dyadic_dim = (dim << dyadic_order_2) + 1

    # Allocate (flattened) PDE grid
    cdef double* pde_grid = <double*>malloc(dyadic_len * dyadic_dim * sizeof(double))
    cdef double deriv, deriv_2

    # Initialization of K array
    for i in range(dyadic_len):
        pde_grid[i * dyadic_dim] = 1.0  # Set K[i, 0] = 1.0

    for j in range(dyadic_dim):
        pde_grid[j] = 1.0  # Set K[0, j] = 1.0

    for j in range(dyadic_dim - 1):
        d = j >> dyadic_order_2
        for i in range(dyadic_len - 1):
            ii = i >> dyadic_order_1

            if lambda_ & (1 << d):
                deriv = beta_frac * (path[ii + 1, d] - path[ii, d])
            else:
                deriv = beta_frac * (path[ii, d] - path[ii + 1, d])

            deriv_2 = deriv * deriv * twelth
            pde_grid[(i + 1) * dyadic_dim + (j + 1)] = (
                    (pde_grid[(i + 1) * dyadic_dim + j] + pde_grid[i * dyadic_dim + (j + 1)])
                    * (1.0 + 0.5 * deriv + deriv_2)
                    - pde_grid[i * dyadic_dim + j] * (1.0 - deriv_2)
            )
    result = <double> pde_grid[dyadic_len * dyadic_dim - 1]

    # Free PDE grid
    free(pde_grid)
    return result

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef double chen_(double[:,:] path) nogil:
    cdef int dim = path.shape[1]
    cdef int path_length = path.shape[0]
    cdef double prod, result
    cdef int i, j, k, m, ii

    cdef double* last_coeffs = <double*>malloc((dim+1) * sizeof(double))
    cdef double* new_coeffs = <double*>malloc((dim+1) * sizeof(double))

    last_coeffs[0] = 1
    new_coeffs[0] = 1

    fact_prod = 1.
    for i in range(1, dim + 1):
        fact_prod *= (path[1, i-1] - path[0, i-1]) / i
        last_coeffs[i] = fact_prod # prod / factorial(i)

    for m in range(1, path_length - 1):
        for ii in range(dim):
            i = ii+1
            new_coeffs[i] = last_coeffs[i]

            fact_prod = 1.
            for k in range(i-1, -1, -1):
                fact_prod *= (path[m + 1, k] - path[m, k]) / (i - k)
                new_coeffs[i] += last_coeffs[k] * fact_prod

        new_coeffs, last_coeffs = last_coeffs, new_coeffs

    result = last_coeffs[dim]

    free(last_coeffs)
    free(new_coeffs)

    return result


import torch.cuda
from numba import cuda
import numba

@cuda.jit('int64(int64,int64)', fastmath = True, device = True)
def lambda_sign(lambda_, dim):
    # Returns 1 if lambda_ has an even number of zeros in its binary expansion and -1 otherwise
    # dim is the length of the binary representation of lambda_
    ones_count = 0
    while lambda_:
        ones_count += lambda_ & 1
        lambda_ >>= 1
    return 1 if (dim - ones_count) % 2 == 0 else -1

def get_alpha(beta, scaling_depth, dim, device):
    # Returns Vandermonde coefficients alpha given beta
    alpha_array = torch.empty(scaling_depth + 1, dtype = torch.float64, device = device)
    sign = -1 if scaling_depth & 1 else 1 #(-1) ** scaling_depth

    for i in range(scaling_depth + 1):
        alph = sign
        alph /= beta[i] ** dim
        for j in range(scaling_depth + 1):
            if j == i:
                continue
            alph *= beta[j] / (beta[i] - beta[j])
        alpha_array[i] = alph
    return alpha_array

#####################################################################
#get_coeff_cuda, get_kernel_cuda allocate the entire PDE grid
#####################################################################
def get_kernel_grid_cuda(pde_grid, path, scaling_depth, dyadic_order_1, dyadic_order_2, dim, max_lambda, dyadic_len, dyadic_dim, full = False):
    # Populates PDE grid K of shape (M + 1, lam_lim, dyadic_len, dyadic_dim + 1)

    # Assign initial values
    pde_grid[:, :, 0, :] = 1.
    pde_grid[:, :, :, 0] = 1.

    # Compute beta and alpha for Vandermonde
    beta = torch.linspace(0.1, 1, scaling_depth + 1, dtype=torch.float64, device='cuda') if scaling_depth > 0 else torch.ones(1, dtype=torch.float64, device='cuda')
    alpha = get_alpha(beta, scaling_depth, dim, 'cuda')

    # Total number of antidiagonals
    num_anti_diag = dyadic_len + dyadic_dim

    # Populate grid
    get_kernel_cuda[(scaling_depth + 1, max_lambda), dyadic_dim,](pde_grid, path, alpha, beta, dim, dyadic_dim, dyadic_len, num_anti_diag, dyadic_order_1, dyadic_order_2, full)

def get_coeff_cuda(path, scaling_depth, dyadic_order_1, dyadic_order_2, full = False):

    path_length = path.shape[0]
    dim = path.shape[1]

    # In the method, lambda represents a component-wise scaling. In the case of a central difference,
    # lambda is a vector in {-1,1}^dim. We choose to represent this vector as an integer, whose binary
    # representation denotes the entries of the vector. For example, 5 = (1,0,1)_2 corresponds to the
    # vector (1, -1, 1). max_lambda here denotes the upper bound for these integers.
    max_lambda = 1 << dim

    # Dyadically refined grid dimensions
    dyadic_len = ((path_length - 1) << dyadic_order_1) + 1
    dyadic_dim = (dim << dyadic_order_2)

    # Allocate PDE grids
    pde_grid = torch.empty((scaling_depth + 1, max_lambda, dyadic_len, dyadic_dim + 1), dtype=torch.float64, device='cuda')

    # Populate PDE grids
    get_kernel_grid_cuda(pde_grid, path, scaling_depth, dyadic_order_1, dyadic_order_2, dim, max_lambda, dyadic_len, dyadic_dim, full)

    # Sum as necessary
    if full:
        result = torch.empty((path_length, dim), dtype=torch.float64, device='cuda')
        for i in range(dim, 0, -1):
            result[:, i - 1] = torch.sum(pde_grid[:, :(1 << i), ::(1 << dyadic_order_1), i << dyadic_order_2], dim=(0, 1)) * (0.5 ** i)
        return result
    else:
        result = torch.sum(pde_grid[:, :, -1, -1])
        return result * (0.5 ** dim)

@cuda.jit('void(float64[:,:,:,:],float64[:,:],float64[:],float64[:],int64,int64,int64,int64,int64,int64,boolean)', fastmath = True)
def get_kernel_cuda(pde_grid, path, alpha_arr, beta_arr, dim, dyadic_dim, dyadic_len, num_anti_diag, dyadic_order_1, dyadic_order_2, full):
    # Each block corresponds to a single (beta, lambda_) pair.
    M_idx = int(cuda.blockIdx.x)
    lambda_ = int(cuda.blockIdx.y)
    # Each thread works on a node of a diagonal.
    thread_id = int(cuda.threadIdx.x)

    beta_frac = beta_arr[M_idx] / (1 << (dyadic_order_1 + dyadic_order_2))
    twelth = 1. / 12

    for p in range(2, num_anti_diag):  # First two antidiagonals are initialised to 1
        start_j = max(1, p - dyadic_len + 1)
        end_j = min(p, dyadic_dim + 1)

        j = start_j + thread_id

        if j < end_j:
            d = (j - 1) >> dyadic_order_2

            i = p - j  # Calculate corresponding i (since i + j = p)
            ii = ((i - 1) >> dyadic_order_1) + 1

            if lambda_ & (1 << d):
                deriv = beta_frac * (path[ii, d] - path[ii - 1, d])
            else:
                deriv = beta_frac * (path[ii - 1, d] - path[ii, d])

            deriv_2 = deriv * deriv * twelth
            pde_grid[M_idx, lambda_, i, j] = (pde_grid[M_idx, lambda_, i, j - 1] + pde_grid[M_idx, lambda_, i - 1, j]) * (
                    1. + 0.5 * deriv + deriv_2) - pde_grid[M_idx, lambda_, i - 1, j - 1] * (1. - deriv_2)

        # Wait for other threads in this block
        cuda.syncthreads()

    #scale as necessary
    if full:
        if thread_id < dim:
            j = thread_id + 1
            dim_idx = j << dyadic_order_2
            fact = lambda_sign(lambda_, j) * alpha_arr[M_idx] * (beta_arr[M_idx] ** (dim - j))
            for i in range(0, dyadic_len, 1 << dyadic_order_1):
                pde_grid[M_idx, lambda_, i, dim_idx] -= 1
                pde_grid[M_idx, lambda_, i, dim_idx] *= fact
    else:
        if thread_id == 0:
            pde_grid[M_idx, lambda_, -1, -1] -= 1
            pde_grid[M_idx, lambda_, -1, -1] *= lambda_sign(lambda_, dim) * alpha_arr[M_idx]


#####################################################################
#get_coeff_cuda_2, get_kernel_cuda_2 allocate only the required 3 anti-diagonals
#These are preferred in the single coefficient case for memory efficiency
#####################################################################
def get_coeff_cuda_2(path, scaling_depth, dyadic_order_1, dyadic_order_2):
    path_length = path.shape[0]
    dim = path.shape[1]

    # In the method, lambda represents a component-wise scaling. In the case of a central difference,
    # lambda is a vector in {-1,1}^dim. We choose to represent this vector as an integer, whose binary
    # representation denotes the entries of the vector. For example, 5 = (1,0,1)_2 corresponds to the
    # vector (1, -1, 1). max_lambda here denotes the upper bound for these integers.
    max_lambda = 1 << dim

    # Dyadically refined grid dimensions
    dyadic_len = ((path_length - 1) << dyadic_order_1) + 1
    dyadic_dim = (dim << dyadic_order_2)

    # Allocate array to store results
    results = torch.empty((scaling_depth + 1, max_lambda), dtype=torch.float64, device ='cuda') #Container for results of kernel evaluations

    # Get beta and alpha for Vandermonde
    beta = torch.linspace(0.1, 1, scaling_depth + 1, dtype=torch.float64, device ='cuda') if scaling_depth > 0 else torch.ones(1, dtype=torch.float64, device ='cuda')
    alpha = get_alpha(beta, scaling_depth, dim, 'cuda')

    #Total number of antidiagonals
    num_anti_diag = dyadic_len + dyadic_dim

    # Populate results
    sharedmem = 24 * (dyadic_dim + 1)
    get_kernel_cuda_2[(scaling_depth + 1, max_lambda), dyadic_dim, 0, sharedmem](results, path, alpha, beta, dim, dyadic_len, dyadic_dim, num_anti_diag, dyadic_order_1, dyadic_order_2)

    # Sum and scale
    result = torch.sum(results)
    return result * (0.5 ** dim)

@cuda.jit('void(float64[:,:],float64[:,:],float64[:],float64[:],int64,int64,int64,int64,int64,int64)', fastmath = True)
def get_kernel_cuda_2(results, path, alpha_arr, beta_arr, dim, dyadic_len, dyadic_dim, n_anti_diag, dyadic_order_1, dyadic_order_2):
    # Each block corresponds to a single (beta, lambda_) pair.
    M_idx = int(cuda.blockIdx.x)
    lambda_ = int(cuda.blockIdx.y)
    # Each thread works on a node of a diagonal.
    thread_id = int(cuda.threadIdx.x)

    beta_frac = beta_arr[M_idx] / (1 << (dyadic_order_1 + dyadic_order_2))
    twelth = 1. / 12

    #Shared memory for the 3 antidiagonals
    shared_memory = cuda.shared.array(shape=0, dtype=numba.float64)

    #Initialise to 1
    for i in range(3):
        shared_memory[i * (dyadic_dim + 1) + thread_id] = 1.

    #Only dyadic_dim many threads passed, so deal with last index using thread 0
    if thread_id == 0:
        for i in range(3):
            shared_memory[i * (dyadic_dim + 1) + dyadic_dim] = 1.

    #Wait for initialisation of shared memory
    cuda.syncthreads()

    # Indices determine the start points of the antidiagonals in memory
    # Instead of swaping memory, we swap indices to avoid memory copy
    prev_prev_diag_idx = 0
    prev_diag_idx = (dyadic_dim + 1)
    next_diag_idx = 2 * (dyadic_dim + 1)

    for p in range(2, n_anti_diag):  # First two antidiagonals are initialised to 1
        start = max(1, p - dyadic_len + 1)
        end = min(p, dyadic_dim + 1)

        j = start + thread_id

        if j < end:
            d = (j - 1) >> dyadic_order_2

            i = p - j  # Calculate corresponding i (since i + j = p)
            ii = ((i - 1) >> dyadic_order_1) + 1

            if lambda_ & (1 << d):
                deriv = beta_frac * (path[ii, d] - path[ii - 1, d])
            else:
                deriv = beta_frac * (path[ii - 1, d] - path[ii, d])

            deriv_2 = deriv * deriv * twelth

            # Update the next diagonal entry
            shared_memory[next_diag_idx + j] = (
                          shared_memory[prev_diag_idx + j] +
                          shared_memory[prev_diag_idx + j - 1]) * (
                              1. + 0.5 * deriv + deriv_2) - shared_memory[prev_prev_diag_idx + j - 1] * (1. - deriv_2)

        # Wait for all threads in this block to finish
        cuda.syncthreads()

        # Rotate the diagonals (swap indices, no data copying)
        prev_prev_diag_idx, prev_diag_idx, next_diag_idx = prev_diag_idx, next_diag_idx, prev_prev_diag_idx

        #Make sure all threads wait for the rotation of diagonals
        cuda.syncthreads()

    #Add results
    if thread_id == 0:
        results[M_idx, lambda_] = lambda_sign(lambda_, dim) * alpha_arr[M_idx] * (shared_memory[prev_diag_idx + dyadic_dim] - 1)

#####################################################################
# A serial implementation on CUDA for an apples-to-apples comparison
# For timing purposes only
#####################################################################
def get_coeff_cuda_serial(path, scaling_depth, dyadic_order_1, dyadic_order_2):
    path_length = path.shape[0]
    dim = path.shape[1]

    # In the method, lambda represents a component-wise scaling. In the case of a central difference,
    # lambda is a vector in {-1,1}^dim. We choose to represent this vector as an integer, whose binary
    # representation denotes the entries of the vector. For example, 5 = (1,0,1)_2 corresponds to the
    # vector (1, -1, 1). max_lambda here denotes the upper bound for these integers.
    max_lambda = 1 << dim
    result = 0

    # Dyadically refined grid dimensions
    dyadic_len = ((path_length - 1) << dyadic_order_1) + 1
    dyadic_dim = (dim << dyadic_order_2) + 1

    beta = torch.linspace(0.1, 1, scaling_depth + 1, dtype=torch.float64, device='cuda') if scaling_depth > 0 else torch.ones(1, dtype=torch.float64, device='cuda')
    alpha = get_alpha(beta, scaling_depth, dim, 'cuda')

    pde_grid = torch.empty((dyadic_len, dyadic_dim), dtype = torch.float64, device = "cuda")

    for lambda_ in range(max_lambda):
        for i in range(scaling_depth + 1):
            get_kernel_cuda_serial[1,1](pde_grid, path, lambda_, beta[i], alpha[i], dyadic_order_1, dyadic_order_2)
            result += pde_grid[-1,-1]

    result *= 0.5 ** dim
    return result

@cuda.jit('void(float64[:,:], float64[:,:],int64,float64,float64,int64,int64)', fastmath = True)
def get_kernel_cuda_serial(pde_grid, path, lambda_, beta, alpha, dyadic_order_1, dyadic_order_2):
    path_length = path.shape[0]
    dim = path.shape[1]
    beta_frac = beta / (1 << (dyadic_order_1 + dyadic_order_2))
    twelth = 1. / 12

    # Dyadically refined grid dimensions
    dyadic_len = ((path_length - 1) << dyadic_order_1) + 1
    dyadic_dim = (dim << dyadic_order_2) + 1

    # Initialization of K array
    for i in range(dyadic_len):
        pde_grid[i, 0] = 1.0

    for j in range(dyadic_dim):
        pde_grid[0, j] = 1.0

    for j in range(dyadic_dim - 1):
        d = j >> dyadic_order_2
        for i in range(dyadic_len - 1):
            ii = i >> dyadic_order_1

            if lambda_ & (1 << d):
                deriv = beta_frac * (path[ii + 1, d] - path[ii, d])
            else:
                deriv = beta_frac * (path[ii, d] - path[ii + 1, d])

            deriv_2 = deriv * deriv * twelth
            pde_grid[i + 1, j + 1] = (
                    (pde_grid[i + 1, j] + pde_grid[i, j + 1])
                    * (1.0 + 0.5 * deriv + deriv_2)
                    - pde_grid[i, j] * (1.0 - deriv_2)
            )
    pde_grid[dyadic_len - 1, dyadic_dim - 1] -= 1
    pde_grid[dyadic_len - 1, dyadic_dim - 1] *= lambda_sign(lambda_, dim) * alpha

#####################################################################
# An implementation of Chen's relation for an apples-to-apples comparison
#####################################################################
def chen_cuda_(path):
    dim = path.shape[1]
    sharedmem = 16 * (dim + 1)
    result = torch.empty(1, dtype = torch.float64, device = "cuda")
    run_chen_cuda[1, dim, 0, sharedmem](path, result)
    return result

@cuda.jit('void(double[:,:], double[:])', fastmath = True)
def run_chen_cuda(path, result):
    thread_id = int(cuda.threadIdx.x)
    dim = path.shape[1]
    path_length = path.shape[0]

    shared_memory = cuda.shared.array(shape=0, dtype=numba.float64)

    last_coeffs_idx = 0
    new_coeffs_idx = dim + 1

    # Set the 0^th coefficient to 1
    if thread_id == 0:
        shared_memory[last_coeffs_idx] = 1
        shared_memory[new_coeffs_idx] = 1

    # Compute coefficients for the first linear segment
    fact_prod = 1.
    for i in range(1, dim + 1):
        fact_prod *= (path[1, i - 1] - path[0, i - 1]) / i
        shared_memory[last_coeffs_idx + i] = fact_prod  # prod / factorial(i)

    for LL in range(1, path_length - 1):
        for ii in range(dim):
            i = ii+1
            shared_memory[new_coeffs_idx + i] = shared_memory[last_coeffs_idx + i]

            fact_prod = 1.
            for k in range(i-1, -1, -1):
                fact_prod *= (path[LL + 1, k] - path[LL, k]) / (i - k)
                shared_memory[new_coeffs_idx + i] += shared_memory[last_coeffs_idx + k] * fact_prod

        new_coeffs_idx, last_coeffs_idx = last_coeffs_idx, new_coeffs_idx

    #Store result
    result[0] = shared_memory[last_coeffs_idx + dim]


from cython_backend import *
from .cuda_backend import *
import torch
import warnings
from numba import NumbaPerformanceWarning

def reorder_path(path, multi_index):
    # Retrieves the path x^{multi_index}
    path_length = path.shape[0]
    new_dim = len(multi_index)
    new_path = torch.empty((path_length, new_dim), dtype=torch.double, device = path.device)
    for i in range(new_dim):
        new_path[:, i] = path[:, multi_index[i]]
    return new_path

def coeff(path, multi_index, dyadic_order = 2, scaling_depth = 2, parallel = True, normalise = True, full = False):
    """
    Computes the signature coefficient S(X)^{multi_index} using the kernel approach.

    :param path: Underlying path. Should be a torch.Tensor of shape (path length, path dimension)
    :param multi_index: Array-like multi-index of signature coefficient to compute. If full = True this is the terminal multi-index in the grid.
    :param dyadic_order: int or 2-tuple. If int, dyadic order is taken to be the same over both dimensions of the PDE grid. Otherwise, dyadic orders taken from tuple (dyadic_order_len, dyadic_order_dim).
    :param scaling_depth: Vandermonde scaling depth.
    :param parallel: If true, computes in parallel when using cpu. This parameter is ignored if X.device = "cuda".
    :param normalise: If true, pre-scales X by max(abs(X)) before computing coefficient. Can improve performance for paths taking large values.
    :param full: If true, returns the full grid of signature coefficients of shape (path length, multi-index length).
    :return: torch.Tensor
    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=NumbaPerformanceWarning)
        return coeff_(path, multi_index, dyadic_order, scaling_depth, parallel, normalise, full)


def coeff_(path, multi_index, dyadic_order = 2, scaling_depth = 2, parallel = True, normalise = True, full = False):
    _multi_index = list(multi_index)
    dim = len(_multi_index)
    path_length = path.shape[0]

    if type(dyadic_order) == int:
        dyadic_order_1, dyadic_order_2 = dyadic_order, dyadic_order
    elif type(dyadic_order) == tuple and len(dyadic_order) == 2:
        dyadic_order_1, dyadic_order_2 = dyadic_order
    else:
        raise ValueError("dyadic_order must by int or tuple of length 2")

    # It is obviously better to return the increment in the dim = 1 case as below,
    # but we ignore this here for consistency of results.
    # if len(_multi_index) == 1:
    #     return X[-1, _multi_index[0]] - X[0, _multi_index[0]]

    if path_length < 2 or dim < 1:
        return torch.tensor(0.)

    path_at_multi_index = reorder_path(path, _multi_index)

    if normalise:
        scaling = torch.max(torch.abs(path))
        path_at_multi_index /= scaling

    if path.device.type == "cpu":
        if not full:
            result = get_coeff_cython(path_at_multi_index.numpy(), scaling_depth, dyadic_order_1, dyadic_order_2, parallel)
        else:
            raise ValueError("full = True not supported with cpu")
    else:
        if full:
            result = get_coeff_cuda(path_at_multi_index.cuda(), scaling_depth, dyadic_order_1, dyadic_order_2, full = True)
        else:
            #get_coeff_cuda_2 does not store the grid so is more memory efficient if we're only interested in one coefficient
            result = get_coeff_cuda_2(path_at_multi_index.cuda(), scaling_depth, dyadic_order_1, dyadic_order_2)

    if normalise:
        if full:
            scaling_pow = float(scaling)
            for i in range(1, dim + 1):
                result[:, i - 1] *= scaling_pow
                scaling_pow *= scaling
        else:
            result *= scaling ** dim

    return result

def coeff_cuda_serial(path, multi_index, dyadic_order=2, scaling_depth=2, normalise=True):
    """
    For timing purposes only. A serial implementation on CUDA for an apples-to-apples comparison.

    :param path: Underlying path. Should be a torch.Tensor of shape (path length, path dimension)
    :param multi_index: Array-like multi-index of signature coefficient to compute. If full = True this is the terminal multi-index in the grid.
    :param dyadic_order: int or 2-tuple. If int, dyadic order is taken to be the same over both dimensions of the PDE grid. Otherwise, dyadic orders taken from tuple (dyadic_order_len, dyadic_order_dim).
    :param scaling_depth: Vandermonde scaling depth.
    :param normalise: If true, pre-scales X by max(abs(X)) before computing coefficient. Can improve performance for paths taking large values.
    :return: torch.Tensor
    """
    if path.device == "cpu":
        raise ValueError("X not on CUDA in cuda_serial")

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=NumbaPerformanceWarning)
        return _coeff_cuda_serial(path, multi_index, dyadic_order, scaling_depth, normalise)

def _coeff_cuda_serial(path, multi_index, dyadic_order=2, scaling_depth=2, normalise=True):
    _multi_index = list(multi_index)
    dim = len(_multi_index)
    path_length = path.shape[0]

    if type(dyadic_order) == int:
        dyadic_order_1, dyadic_order_2 = dyadic_order, dyadic_order
    elif type(dyadic_order) == tuple and len(dyadic_order) == 2:
        dyadic_order_1, dyadic_order_2 = dyadic_order
    else:
        raise ValueError("dyadic_order must by int or tuple of length 2")

    # It is obviously better to return the increment in the dim = 1 case as below,
    # but we ignore this here for consistency of results.
    # if len(_multi_index) == 1:
    #     return X[-1, _multi_index[0]] - X[0, _multi_index[0]]

    if path_length < 2 or path.shape[1] < 1:
        return torch.tensor(0.)

    path_at_multi_index = reorder_path(path, _multi_index)

    if normalise:
        scaling = torch.max(torch.abs(path))
        path_at_multi_index /= scaling

    result = get_coeff_cuda_serial(path_at_multi_index.cuda(), scaling_depth, dyadic_order_1, dyadic_order_2)

    if normalise:
        result *= scaling ** dim

    return result

def coeff_chen_cython(path, multi_index):
    """
    Computes the signature coefficient S(X)^{multi_index} using Chen's relation, computed with Cython.

    :param path: Underlying path. Should be a torch.Tensor of shape (path length, path dimension)
    :param multi_index: Array-like multi-index of signature coefficient to compute
    :return: double
    """
    _multi_index = list(multi_index)
    path_at_multi_index = reorder_path(path, _multi_index)
    return chen_(path_at_multi_index.numpy())

def coeff_chen_cuda(path, multi_index):
    """
    Computes the signature coefficient S(X)^{multi_index} using Chen's relation, computed with CUDA.
    CUDA is used purely for comparability with coeff(x.cuda(), multi_index) and Chen's relation is run serially on one thread.

    :param path: Underlying path. Should be a torch.Tensor of shape (path length, path dimension)
    :param multi_index: Array-like multi-index of signature coefficient to compute
    :return: double
    """

    _multi_index = list(multi_index)
    path_at_multi_index = reorder_path(path, _multi_index)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=NumbaPerformanceWarning)
        result = chen_cuda_(path_at_multi_index).cpu()

    return float(result)


import os.path
from Cython.Build import cythonize
from setuptools import setup, Extension
import numpy as np

this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

ext_modules = [
    Extension(
        name="cython_backend",
        sources=["sigcoeff/cython_backend.pyx"],
        include_dirs = [".", np.get_include()],
        extra_compile_args=['-openmp'],
        extra_link_args=['-openmp'],
    )
]

for e in ext_modules:
    e.cython_directives = {'language_level': "3"}

setup(
    name="sigcoeff",
    version="1.0.1",
    description="Sparse signature coefficient computations via kernels",
    packages=["sigcoeff"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Win32 (MS Windows)",
        "Intended Audience :: Developers",
        "Intended Audience :: Financial and Insurance Industry",
        "Intended Audience :: Healthcare Industry",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: Unix",
        "Programming Language :: Python",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Mathematics",
    ],
    url="https://github.com/daniil-shmelev/sigcoeff",
    author="Daniil Shmelev",
    author_email="daniil.shmelev23@imperial.ac.uk",
    setup_requires=["cython"],
    install_requires=["torch>=2.4.1",
    "numba>=0.60.0",
    "cython>=3.0.11",
    "numpy>=1.26.4",
    "setuptools>=74.1.2"
    ],
    extras_require={
        "dev": [
            "check-manifest",
            "twine",
            "black",
        ],
    },
    ext_modules=cythonize(ext_modules),
    package_data = {"sigcoeff": ["*.pyx", "*.pxd", "*.so", "*.pyd"]},
    include_package_data=True
)